#include <string>

#include "analyze_new.h"

using std::string;

//////////////////////////////////////
// helper classes

struct CMinMax
{
  int m_iMin;
  int m_iMax;
  double m_dMin;
  double m_dMax;
  double m_dh;

  CMinMax(int iMax, double dh) :
    m_iMin(iMax), m_iMax(0),
    m_dMin(iMax), m_dMax(0),
    m_dh(dh)
  {}


  double Interpol(double h0, double h1, double href) {
    if (h1-h0 > 1e-10) {
      double retval;
      retval= (href-h0)/(h1-h0);
      if( retval < 0 )
        retval= 0;
      return (retval> 1.0? 1.0 : retval);
    }
    else
      return 0.5;
  }


  void operator()(int i, double dl, double d, double dr)
  {
    if (i<=m_iMin) {
      m_iMin = i;
      if( m_dMin > i-1+Interpol(dl,d,m_dh) )
        m_dMin = i-1+Interpol(dl,d,m_dh);
    }

    if (i>=m_iMax) {
      m_iMax = i;
      if( m_iMax < i+Interpol(d,dr,m_dh) )
        m_dMax = i+Interpol(d,dr,m_dh);
    }
  }

  void Check(const char* pszName)
  {
    if (m_dMin > m_dMax) {
      double d = m_dMin;
      m_dMin = m_dMax;
      m_dMax = d;
      cout << pszName << ": Min/max detection failed!" << endl;
    }
  }

  int Center() {
    return (m_iMin+m_iMax)/2;
  }

};


//////////////////////////////////////
// analyze
//

analyze::analyze(const dunepar& P)
{
  sep_angle= P.getdefault("sep.angle", 30.0);
  std::string strFile = P.getdefault<string>("save.dir", "./");
  if (strFile[strFile.length()-1] != '/') {
    strFile += '/';
  }
  strFile += "time.dat";
  m_os.open(strFile.c_str());

  if (!m_os) {
    cout << "Open file \"" << strFile << "\" failed!" << endl;
    exit(1);
  }

  m_os << "%# 1: iteration n\n"
       << "%# 2: time t\n"
       << "%# 3: height h\n"
       << "%# 4: width w\n"
       << "%# 5: length l\n"
       << "%# 6: horn length l_h\n"
       << "%# 7: slip-face length l_s\n"
       << "%# 8: windward side length l_0\n"
       << "%# 9: volume V\n"
       << "%#10: distance traveled by the dune in X\n"
       << "%#11: maximum dhdt\n"
       << "%#12: dune influx\n"
       << "%#13: dune outflux\n"
       << "%#14: time step for dh/dt\n"
       << "%#15: half mean length\n"
       << "%#16: mean sand flux over the field\n"
       << "%#17: mean vegetation density\n"
       << "%#18: position of the slip face's foot\n"
       << endl;
}


void analyze::Calc(int t, double timestep, double shift_dist_x, double maxchange, double halfLength,
	double qin, double qout, double meanFlux, double meanVegetRho, const TFktScal& m_h)
{
  const double dx = duneglobals::dx();

  m_dTime += timestep;

  // calc dune properties

  const double dHMax = m_h.GetMax();
  const double dHRef = dHMax/100.;

  double minSlipos = 0;
  for (int x=m_h.SizeX()-1; x > 0; x--) 
  	if(m_h(x,m_h.SizeY()/2) > 0.1){
		minSlipos = x;
		break;
  	}
	
  CMinMax mmX(m_h.SizeX(), dHRef);
  CMinMax mmY(m_h.SizeY(), dHRef);

  // scan surface at level dHRef -> length and width

  for (int y=1; y<m_h.SizeY()-1; y++) {
    for (int x=1; x<m_h.SizeX()-1; x++) {
      if (m_h(x,y) > dHRef) {
	mmX(x, m_h(x-1,y), m_h(x,y), m_h(x+1,y));
	mmY(y, m_h(x,y-1), m_h(x,y), m_h(x,y+1));
      }
    }
  }
  mmX.Check("Length L");
  mmY.Check("Width W");


  // scan symmetry line -> l_h, l_0

  CMinMax mmXSym(m_h.SizeX(), dHRef);

  const int iYSym = mmY.Center();
  for (int x=1; x<m_h.SizeX()-1; x++) {
    if (m_h(x,iYSym) > dHRef) {
      mmXSym(x, m_h(x-1,iYSym), m_h(x,iYSym), m_h(x+1,iYSym));
    }
  }
  mmXSym.Check("Symmetry line L0+Ls");


  // calc position of max height, flux, shear

  double dMaxH = 0.;
  int iXH(0);
  for (int x=mmXSym.m_iMin+1; x<mmXSym.m_iMax-2; x++) 
  {
    if (dMaxH < m_h(x,iYSym)) {
      dMaxH = m_h(x,iYSym);
      iXH = x;
    }
  }
  double dXH = GetMaxPos(m_h(iXH-1,iYSym), m_h(iXH,iYSym), m_h(iXH+1,iYSym));
  dXH = ((double)iXH + dXH) * dx;

  // scan for the crest position

  int iXBrink = -1;
  double m_sepslope = tan(M_PI*sep_angle/180.0);
  for (int x=mmXSym.m_iMin; x<mmXSym.m_iMax-1; x++) {
    if (m_h(x,iYSym)-m_h(x-1,iYSym) > -m_sepslope*dx && m_h(x+1,iYSym)-m_h(x,iYSym) < -m_sepslope*dx) {
      iXBrink = x;
      break;
    }
  }

  if (iXBrink == -1) {
    // use maximum instead / no slip-face
    double dMax = 0.;
    for (int x=mmXSym.m_iMin; x<mmXSym.m_iMax-1; x++) {
      if (dMax < m_h(x,iYSym)) {
	dMax = m_h(x,iYSym);
	iXBrink = x;
      }
    }
  }

  if (iXBrink == -1) {
    iXBrink = (mmXSym.m_iMin+mmXSym.m_iMax)/2;
    if (m_bMsg) {
      cout << "--> Brink/crest detection failed!" << endl;
    }
  }

  // ---- calc lengths ----

  m_dCenter = (mmX.m_dMax+mmX.m_dMin)*dx/2;

  const double dL = (mmX.m_dMax-mmX.m_dMin)*dx;
  const double dW = (mmY.m_dMax-mmY.m_dMin)*dx;
  const double dL0 = (iXBrink-mmXSym.m_dMin)*dx;
  const double dLs = (mmXSym.m_dMax-iXBrink)*dx;
  const double dLh = (mmX.m_dMax-mmXSym.m_dMax)*dx;

  // ---- volume / mass ----

  const double dVol= m_h.Integrate();
  
  // ---- write data ----

  m_os << t << " "                    // 1: iterations
       << m_dTime << " "              // 2: time in s
       << dHMax << " "                // 3: height
       << dW << " "                   // 4: width w
       << dL << " "                   // 5: length l
       << dLh << " "                  // 6: horn length l_h
       << dLs << " "                  // 7: slip-face length l_s
       << dL0 << " "                  // 8: windward side length l_0
       << dVol << " "                 // 9: volume / mass of sand
       << shift_dist_x << " "	      // 10: distance traveled by the dune in X
       << maxchange << " "	      // 11: maximum dhdt
       << qin << " "                  // 12: dune in flux
       << qout << " "                 // 13: dune out flux
       << timestep << " "             // 14: time step for dh/dt
       << halfLength << " "	      // 15: half mean length
       << meanFlux << " "	// 16: mean sand flux over the field
       << meanVegetRho << " "	// 17: mean vegetation density
       << minSlipos*dx << " "	// 18: position of the slip face's foot
       << endl;
}


double analyze::GetMaxPos(double f0, double f1, double f2)
{
  // assert that: f0 < f1 and f1 > f2
  if (f0 > f1 || f2 > f1)
    return 0;

  // calc derivatives: d1 > 0, d2 < 0
  double d1 = f1-f0;
  double d2 = f2-f1;

  // calc zero pos.  d1 + epsilon * (d2-d1) = 0
  double epsilon;
  if (d1 != d2) {
    epsilon = d1 / (d1-d2);
  } else {
    epsilon = 0.5;
  }
  return epsilon - 0.5;
}
