/******************************************************************************
    $Id: flux_stationary.cc,v 1.27 2005/04/13 18:33:12 duran Exp $
******************************************************************************/

#include <stdlib.h>
#include <float.h>
#include <math.h>
#include <iostream>

#include "physics_const.h"
#include "globals.h"

#include "flux_stationary.h"


//*****************************************************************************
//  class flux3d_stationary

/*!  Constructor: reads relevant parameters and creates auxiliary arrays for
  the velocity and the density of entrained sand rho and its gradient.  */

flux3d_stationary::flux3d_stationary( const dunepar& parameters ): dunedata(parameters)
{
//  Remove these comments after we have truly separated 2d and 3d parts of the
//  software.  At the moment, this class might still be used for "2d" 3-by-x
//  simulations.
//	if( !duneglobals::sim3d() ) {
//	  cerr << "flux3d_stationary Constructor: FATAL ERROR: This class can only be used for three-dimensional simulations!" << endl;
//	  exit(1);
//	}

  m_rho.Create( duneglobals::nx(), duneglobals::ny(), duneglobals::dx() );
  m_rho.SetAll(0.0);

  m_TempScal.Create( duneglobals::nx(), duneglobals::ny(), duneglobals::dx() );
  m_TempScal.SetAll(0.0);

  m_u.Create( duneglobals::nx(), duneglobals::ny(), duneglobals::dx() );

  m_x_periodic= duneglobals::periodic_x();
  m_y_periodic= duneglobals::periodic_y();
  
  // Get the parameters:
  m_d_grain= parameters.getrequired<double>( "salt.d" );
  m_fluid_viscosity= parameters.getrequired<double>( "fluid_viscosity" );
  double fluid_viscosity_kin = m_fluid_viscosity / duneglobals::rho_fluid();
  m_g= parameters.getdefault( "salt.g", physics_constants::g );
  
  double s = duneglobals::rho_grains()/duneglobals::rho_fluid();
  double Aft = parameters.getrequired<double>( "Bagnold-shields_parameter" );
  double At = 0.8 * Aft;
  m_u_star_ft = Aft * sqrt(m_g*m_d_grain*(s-1.));
  m_u_star_t = (At/Aft) * m_u_star_ft;
  cout << "flux3d_stationary constructor: u*ft = " << m_u_star_ft << "m/s has been calculated from Bagnold-Shields Parameter\n"; 
  cout << "flux3d_stationary constructor: u*t = " << m_u_star_t << "m/s (= ratio_u*t/u*ft multiplied by u*ft)\n"; 
  
  double r= parameters.getdefault( "salt.Z/zm", 0.);
  
  /*! Auxiliar quantities */
  double t_v = exp(1./3. * log(fluid_viscosity_kin / (m_g * m_g)));	//characteristic time
  double l_v = exp(1./3. * log(fluid_viscosity_kin * fluid_viscosity_kin / (At*At * m_g*(s-1))));	//characteristic lenght
  double S = 0.25 * m_d_grain / fluid_viscosity_kin * sqrt(m_g*m_d_grain*(s-1.));
  
  // Parameters
  m_C_diff= parameters.getdefault( "salt.D", 0.0 );
  m_beta= parameters.getrequired<double>( "beta" );
  m_gamma= parameters.getrequired<double>( "gamma" );
  
  m_z0= m_d_grain / 20.;
  cout << "flux3d_stationary constructor: z0 = " << m_z0 << endl;
  m_zm= 14./(1.+1.4*r) * m_u_star_t * t_v;
  cout << "flux3d_stationary constructor: zm = " << m_zm << endl;
  m_z1= 35. * l_v;
  cout << "flux3d_stationary constructor: z1 = " << m_z1 << endl;
  m_alpha= 0.18 * (m_d_grain / l_v);
  cout << "flux3d_stationary constructor: alpha = " << m_alpha << endl;
  
  double Ad = 0.95, Bd = 5.12; //natural sand
  m_C_drag = 4./3. * (Ad + sqrt(2 * m_alpha) * Bd / S)*(Ad + sqrt(2 * m_alpha) * Bd / S);
  cout << "flux3d_stationary constructor: Cd = " << m_C_drag << endl;
  
  m_DeltaU = sqrt(4.0/3.0 * m_g * (s-1) * m_d_grain / (m_C_drag * 2 * m_alpha));
  
  // derived quantities:
  double z0zm= m_z0/m_zm;
  double z1zm= m_z1/m_zm;
  m_log_z1z0= log(m_z1/m_z0);
  
  if( r < z0zm )	r = z0zm;
  double log_r = log(r/z0zm);
  cout << "flux3d_stationary constructor: Z/zm = " << r << endl; 
  
  double b_inf = -0.5772 - log(z0zm) + z0zm;
  double b_Z = (r < 2. ? log_r-(r - z0zm)+0.25*(r*r - z0zm*z0zm)-0.042*(r*r*r - z0zm*z0zm*z0zm) : b_inf - exp(-r)/r);
  m_M = (log_r > 0 ? log_r/b_Z : 1.);
  cout << "flux3d_stationary constructor: M = " << m_M << endl;
  
  m_b_z1 = (z1zm < 2. ? m_log_z1z0-(z1zm - z0zm)+0.25*(z1zm*z1zm - z0zm*z0zm)-0.042*(z1zm*z1zm*z1zm - z0zm*z0zm*z0zm) : b_inf - exp(-z1zm)/z1zm);
  m_Cd_dx= m_C_diff / (duneglobals::dx()*duneglobals::dx());
  m_2alpha_g= 2.0*m_alpha / m_g;
  m_tau_t= duneglobals::rho_fluid() * m_u_star_t * m_u_star_t;
  m_gamma_2alpha_g2= m_gamma / (m_2alpha_g * m_2alpha_g);
  m_tau_ft= duneglobals::rho_fluid() * m_u_star_ft * m_u_star_ft;
  m_beta_2alpha_g_tau_ft = m_beta / (m_2alpha_g * m_tau_ft);
}


/*!  The destructor currently does nothing.  */

flux3d_stationary::~flux3d_stationary()
{
}

double flux3d_stationary::Satflux_upwind(double u_star)
{
  double u_s = 2.5 *u_star *(m_log_z1z0 - (1.- u_star_at(u_star)/u_star)*m_b_z1) - m_DeltaU;
  double flux = 2*m_alpha/m_g * m_tau_t*((u_star/m_u_star_t)*(u_star/m_u_star_t) - 1) * u_s;
  return flux;
}

double flux3d_stationary::SatLength_upwind(double u_star)
{
  double u_s = 2.5 *u_star *(m_log_z1z0 - (1.- u_star_at(u_star)/u_star)*m_b_z1) - m_DeltaU;
  double L_sat = 2*m_alpha/m_g * u_s*u_s /(m_gamma * ((u_star/m_u_star_t)*(u_star/m_u_star_t) - 1));
  return L_sat;
}

double flux3d_stationary::u_star_at( double u_star )
{
  double u_t = m_M * m_u_star_t - (m_M - 1.) * u_star;
  return u_t;//(u_t > 0 ? u_t : 0);
}

/*!  Calculates the effective wind velocity.  The drag force under this
  velocity is the mean force affecting the sand grains in the air.  The
  effective velocity is the wind speed at the height m_z1.
  
  In addition to the simplifications used for all this class, this calculation
  neglects the influence of even spatial transient effects on v_eff since they
  were seen to be small (Veit Schwaemmle, Diploma Thesis).  

  V_eff is calculated according to the formula:
  
  v_eff = tau/|tau| * ( sqrt( [4/kappa^2 * z1/zm * 1/rho_fluid] * tau
          + [(1 - z1/zm) * u_star_t^2] ) + [(log(z1/z0) - 2) * u_star_t/kappa )
  
  This formula is equivalent to equation 3.27 in Schwaemmle's thesis.  The
  three expressions in square brackets are computed in the constructor
  (m_veff_sqrt_slope, m_veff_sqrt_offset and m_veff_offset, respectively).  */

void flux3d_stationary::calc_u( double *u_x, double *u_y,
			const double& tau_x, const double& tau_y, double gradh_x, double gradh_y )
{
    double abs_tau = sqrt( tau_x*tau_x + tau_y*tau_y );
    double u_star = sqrt(abs_tau/duneglobals::rho_fluid());
    double u_star_a_t = u_star_at(u_star);
    if(abs_tau <= m_tau_t) {
      *u_x= *u_y= 0.0;
      return;
    }

    //------efective wind velocity-----------

    double abs_u_wind = 2.5 *u_star *(m_log_z1z0 - (1.- u_star_a_t/u_star)*m_b_z1);

    if(abs_u_wind < m_DeltaU){
        *u_x= *u_y= 0.0;
	return;
    }

    double windx= (abs_u_wind/m_DeltaU)*tau_x/abs_tau;
    double windy= (abs_u_wind/m_DeltaU)*tau_y/abs_tau;
    double Dhx= 2.*m_alpha*gradh_x;
    double Dhy= 2.*m_alpha*gradh_y;
    double Gx, Gy, G, Fx, Fy, F, u, du, delta_du= 1.0;

    double dux= tau_x/abs_tau+Dhx;
    double duy= tau_y/abs_tau+Dhy;
    du= sqrt(dux*dux+duy*duy);
    du= sqrt(du);
    int j=0;
    while(fabs(delta_du) > 1e-3){
       delta_du= du;
       Fx= du*windx-Dhx;
       Fy= du*windy-Dhy;
       F= sqrt(Fx*Fx+Fy*Fy);
       if(du>0) u = (F-1.0)/du;
       else u=0;
       Gx= u*Dhx+windx;
       Gy= u*Dhy+windy;
       G= sqrt(Gx*Gx+Gy*Gy);
       if(u>0) du= 0.5*(sqrt(1.0+4.0*u*G)-1.0)/u;
       else break;

       delta_du-=du;
       delta_du= delta_du/du;
    }
    double ux, uy;
    ux= u*Fx/F;
    uy= u*Fy/F;
    //ux= windx - du;
    //uy= windy - du;
    if( (ux*tau_x+uy*tau_y) > 0){
      *u_x= ux*m_DeltaU;
      *u_y= uy*m_DeltaU;
    }else{
      *u_x= *u_y= 0;
    }

}

/*!  Computes the sand flux.  To give the initial condition, the column x=0 of
  flux has to be filled with the influx (which will be projected onto the
  direction of the air at those points).  */

void flux3d_stationary::calc( TFktVec& flux, const TFktScal& h, const TFktScal&
h_nonerod, const TFktVec& tau, TFktScal& gamma )
{
  double gradh_x, gradh_y, qin, qout;
  int x, y, prevx, nextx, prevy, nexty;

  for( x= 0; x< duneglobals::nx(); ++x ) {
    if( !x )	{ prevx= 0; nextx= 1; }
    else if( x==duneglobals::nx()-1 ) { prevx= x-1; nextx= x; }
    else	{ prevx= x-1; nextx= x+1; }
    for( y= 0; y< duneglobals::ny(); ++y ) {
      gradh_x= (h(nextx, y) - h(prevx, y))/((nextx-prevx)*duneglobals::dx());
     
      prevy = (y==0)?((m_y_periodic)?(duneglobals::ny()-1):0):(y-1);
      nexty = (y==duneglobals::ny()-1)?((m_y_periodic)?0:y):(y+1);
      
      //gradh_y= (nexty==prevy? 0: ((h(x,nexty)-h(x,prevy))/((nexty-prevy)*duneglobals::dx())));
	  gradh_y= (nexty==prevy? 0: ((h(x,nexty)-h(x,prevy))/((2.)*duneglobals::dx())));
      calc_u( &m_u(x, y)[0], &m_u(x, y)[1], tau(x, y)[0], tau(x, y)[1], gradh_x, gradh_y );
    }
  }

  // now compute the influx in the direction of the shear velocity and the
  // initial sand density.
  for( y= 0; y< duneglobals::ny(); ++y )
    if( fabs(m_u(0, y)[0]) > 0 ) {
      m_rho(0, y)= flux(0,y)[0]/m_u(0,y)[0];
      if( !finite(m_rho(0, y)) )
    	m_rho(0, y)= 0.0;   // very small u_abs might cause the result infinity
    }
    else
      m_rho(0, y)= 0.0;
  

  calc_rho( h, h_nonerod, tau, gamma);
 
  

  // Flux= density * saturated velocity
  for( x= 1; x< duneglobals::nx(); ++x )
    for( y= 0; y< duneglobals::ny(); ++y ) {
      flux(x, y)[0] = m_rho(x, y) * m_u(x, y)[0];
      flux(x, y)[1] = m_rho(x, y) * m_u(x, y)[1];
    }

  for( y= 0; y< duneglobals::ny(); ++y ){
    for( x= 0; x< duneglobals::nx(); ++x ){
      if(gamma(x, y)== -1.0) {
	prevy = (y==0)?((m_y_periodic)?(duneglobals::ny()-1):0):(y-1);
	nexty = (y==duneglobals::ny()-1)?((m_y_periodic)?0:y):(y+1);
	qout= fabs(flux(x,y)[0]) + fabs(flux(x,y)[1]);
	qin = ((x > 0 && flux(x-1,y)[0] > 0) ? flux(x-1,y)[0] : 0) +
	  ((x < duneglobals::nx()-1 && flux(x+1,y)[0] < 0) ? -flux(x+1,y)[0] : 0) +       
	  ((flux(x,prevy)[1] > 0) ? flux(x,prevy)[1] : 0) +
	  ((flux(x,nexty)[1] < 0) ? -flux(x,nexty)[1] : 0);
	gamma(x, y)= (qout - qin) / duneglobals::dx();
      }      
    }
  }
	 
}
/*!  Computes the saltating sand density.  The erosion/deposition rate gamma is
  not computed from the density rho but determined separately while computing
  rho.  */

void flux3d_stationary::calc_rho( const TFktScal& h, const TFktScal& h_nonerod,
const TFktVec& tau, TFktScal& gamma )
{
  double A, B;
  
  double tau_abs, u_abs, div_u, splash;
  int x, y, miny, maxy, prevy, nexty;
  bool yascend, firsty;
  
  for(y=0;y<duneglobals::ny();++y) {
      m_TempScal(0,y) = m_rho(0,y);
  }
  for( x= 1; x< duneglobals::nx(); ++x )
  {
   
    maxy= -1;

    while( true)
    {
      y= maxy+1;
      if( y >= duneglobals::ny() )	break;

      for( ; y< duneglobals::ny() && m_u(x, y)[0]==0.0 && m_u(x, y)[1]==0.0; ++y )
	{  m_rho(x, y)= 0.0; gamma(x,y)= -1.0; m_TempScal(x,y)=0.0;}
      if( y >= duneglobals::ny() )	break;
      miny= y;
      for( ; y< duneglobals::ny() && m_u(x, y)[1]==0 && m_u(x, y)[0]!=0; ++y );
      if( m_u(x, y)[1] < 0 ) {
  	yascend= false;
	while( y< duneglobals::ny() &&
		(m_u(x, y)[1] < 0 || (m_u(x, y)[1]==0 && m_u(x, y)[0]!=0)) )
  	  ++y;
  	--y;
	maxy= y;
      }
      else {
  	yascend= true;
	while( y< duneglobals::ny() &&
		(m_u(x, y)[1] > 0 || (m_u(x, y)[1]==0 && m_u(x, y)[0]!=0)) )
  	  ++y;
  	maxy= y-1;
	y= miny;
      }
      for(/* firsty= true*/; yascend? y<=maxy : y>=miny;/* firsty= false,*/
						    yascend? ++y: --y )
      {
	if( x==duneglobals::nx()-1 || m_u(x+1, y)[0] == 0.0)
    	  div_u= m_u(x, y)[0] - m_u(x-1, y)[0];
        else
    	  div_u= 0.5 * (m_u(x+1, y)[0] - m_u(x-1, y)[0]);
	
	prevy = (y==0)?((m_y_periodic)?(duneglobals::ny()-1):0):(y-1);
	nexty = (y==duneglobals::ny()-1)?((m_y_periodic)?0:y):(y+1);
      
	div_u += 0.5 * (m_u(x,nexty)[1] - m_u(x,prevy)[1]);

	A= m_u(x, y)[0] + div_u;

	B= m_u(x, y)[0] * m_TempScal(x-1, y);
	  if( yascend ) {
	    A += m_u(x, y)[1];
	    B += m_u(x, y)[1] * m_TempScal(x, prevy);
	  }
	  else {
	    A += - m_u(x, y)[1]; 
	    B += -m_u(x, y)[1] * m_TempScal(x,nexty);	   	   
	  }
	A /= duneglobals::dx();
	B /= duneglobals::dx();
	//  Erosion can happen only if the sand layer is thick enough.
	splash= (h(x, y) - h_nonerod(x, y)) / m_d_grain;
      	if( splash > 1.0 )
      	  splash= 1.0;
      	else if( splash < 0.001*m_d_grain )
	  //  	  splash= 0.001*m_d_grain;
	 splash = 0;
	//  Diffusion term:
	if(splash < 1.0 && fabs(m_u(x, y)[1])<1e-1){
		if( x>1 ) {
			A -= m_Cd_dx;
		
			B += m_Cd_dx * (m_TempScal(x-2, y) - 2.0 * m_TempScal(x-1, y));
		}
		
		prevy = (y==0)?((m_y_periodic)?(duneglobals::ny()-1):0):(y-1);
		nexty = (y==duneglobals::ny()-1)?((m_y_periodic)?0:y):(y+1);
		
		B += m_Cd_dx * (0.5 *(m_TempScal(x-1, nexty) + m_TempScal(x-1, prevy))-m_TempScal(x-1,y));
		
	}
        u_abs= sqrt( m_u(x, y)[0]*m_u(x, y)[0] + m_u(x, y)[1]*m_u(x, y)[1] );
  	tau_abs= sqrt( tau(x, y)[0]*tau(x, y)[0] + tau(x, y)[1]*tau(x, y)[1] );

	  if(B < 0 || A <= 0 || !u_abs){
	    m_rho(x,y)= 0.0;
	    gamma(x,y)= -1.0;
	  }else{
	    m_rho(x,y) = solve_for_rho(u_abs, tau_abs, splash, A, B);
	    
	    gamma(x,y)= A*m_rho(x,y)-B;
	  }
	  
      }
    }  
     for(y=0;y<duneglobals::ny();++y) {
      m_TempScal(x,y) = m_rho(x,y);
    }  
  }
    
  
  
}


/*!  Determines rho from a quadratic equation which depends on the
  circumstances.  */
          
double flux3d_stationary::solve_for_rho(const double u,
 const double tau, const double splash, const double A, const double B)
  {
    double rho, rho_s, rho_fs, A0_1, A0, A2, A1, rho_dep, rho_direct, rho_eros;
    double u_star = sqrt(tau/duneglobals::rho_fluid());
    double u_star_a_t = u_star_at(u_star);
    double tau_t = duneglobals::rho_fluid()*u_star_a_t*u_star_a_t;
    // Saturated rho
    rho_s= (tau > m_tau_t ? (tau - tau_t) * m_2alpha_g : 0);
    // Saturated rho for direct entrainment
    rho_fs= (tau > m_tau_ft ? (tau - m_tau_ft) * m_2alpha_g : 0);
    // ---Auxiliar const
    // for aeolian entrainment
    A0_1= m_beta_2alpha_g_tau_ft * splash;
    A0= A0_1 * rho_fs;
    // for aeolian deposition
    A2= m_gamma_2alpha_g2 / (u * tau_t);
    A1= A2 * rho_s;
    
    // deposition:
    // calculation of rho with the assumtion that rho > rho_sat
    rho_dep=(B==0 ? 0 : 0.5 * (sqrt((A - A1)*(A - A1) + 4* A2 * B) - (A - A1)) / A2);
    // verifying the assumtion
    if( rho_dep < rho_s ){
        // erosion:
	if( splash > 0 ){
            // calculation of rho without direct entrainment ( rho > rho_fs)
	    A2*= splash;
	    A1= A2 * rho_s;
	    rho_eros=(B==0 ? 0 : 0.5 * (sqrt((A - A1)*(A - A1) + 4* A2 * B) - (A - A1)) / A2);

	    // verifying the assumtion
	    if( rho_eros < rho_fs ){
	        // calculation of rho with direct entrainment ( rho < rho_fs)
		A1-= A0_1;
		rho_direct= 0.5 * (sqrt((A - A1)*(A - A1) + 4* A2 * (B + A0)) - (A - A1)) / A2;

		// verifying the assumtion
		if( rho_eros > rho_fs )    cout << "no solution for rho!!!!!" << endl;
		rho= rho_direct;
	    }else	rho= rho_eros;
	}else	rho= B / A;
    }else	rho= rho_dep;

    if(rho > 0) return rho;
    else return 0;
}


/*!  Saves the arrays m_u and m_rho.  */

void flux3d_stationary::save_arrays()
{
  save_2d_vecarray( "u", m_u );
  save_2d_scalarray( "rho", m_rho );
}

